
import asyncio
import os
from aiosmb.dcerpc.v5.common.connection.smbdcefactory import SMBDCEFactory
from aiosmb.dcerpc.v5 import bkrp
from aiosmb.wintypes.ntstatus import NTStatus
from aiosmb.dcerpc.v5.rpcrt import DCERPCException
from aiosmb import logger
from aiosmb.commons.utils.extb import pprint_exc
from aiosmb.commons.utils.decorators import red_gen, red, rr
from aiosmb.dcerpc.v5.dtypes import NULL
from aiosmb.dcerpc.v5.rpcrt import RPC_C_AUTHN_LEVEL_NONE, RPC_C_AUTHN_LEVEL_PKT_INTEGRITY, RPC_C_AUTHN_LEVEL_PKT_PRIVACY, DCERPCException, RPC_C_AUTHN_GSS_NEGOTIATE, RPC_C_AUTHN_LEVEL_CONNECT

class SMBBKRP:
	def __init__(self, connection):
		self.connection = connection
		self.dce = None
		
	async def __aenter__(self):
		return self
		
	async def __aexit__(self, exc_type, exc, traceback):
		await self.close()
		return True,None
	
	@red
	async def close(self):
		if self.dce:
			try:
				await self.dce.disconnect()
			except:
				pass
			return
		
		return True,None

	@red
	async def connect(self, open = True):
		rpctransport = SMBDCEFactory(self.connection, filename=r'\protected_storage')
		self.dce = rpctransport.get_dce_rpc()
		self.dce.set_auth_level(RPC_C_AUTHN_LEVEL_PKT_PRIVACY)
		_, err = await self.dce.connect()
		if err is not None:
			return False, err
		_, err = await self.dce.bind(bkrp.MSRPC_UUID_BKRP)
		if err is not None:
			return False, err
		return True,None

	async def retrieve_backup_key(self):
		# Requests the public key part of the server's ClientWrap key pair.
		resp, err = await bkrp.hBackuprKey(self.dce, bkrp.BACKUPKEY_RETRIEVE_BACKUP_KEY_GUID, NULL)
		if err is not None:
			return None, err
		return b''.join(resp['ppDataOut']), None

	async def encrypt_data(self, data):
		resp, err = await bkrp.hBackuprKey(self.dce, bkrp.BACKUPKEY_BACKUP_GUID, data)
		if err is not None:
			return None, err
		return b''.join(resp['ppDataOut']), None

	async def decrypt_data(self, data):
		resp, err = await bkrp.hBackuprKey(self.dce, bkrp.BACKUPKEY_RESTORE_GUID, data)
		if err is not None:
			return None, err
		return b''.join(resp['ppDataOut']), None

	async def decrypt_data_2k(self, data):
		resp, err = await bkrp.hBackuprKey(self.dce, bkrp.BACKUPKEY_RESTORE_GUID_WIN2K, data)
		if err is not None:
			return None, err
		return b''.join(resp['ppDataOut']), None


async def amain(url):
	import traceback
	import hashlib
	from aiosmb.commons.connection.url import SMBConnectionURL
	from aiosmb.commons.interfaces.machine import SMBMachine

	url = SMBConnectionURL(url)
	connection = url.get_connection()
	_, err = await connection.login()
	if err is not None:
		print(err)
		raise err
	
	async with SMBBKRP(connection) as b:
		_, err = await b.connect()
		if err is not None:
			print(err)
			print(traceback.format_tb(err.__traceback__))
			return
		
		res, err = await b.retrieve_backup_key()
		if err is not None:
			print(err)
			print(traceback.format_tb(err.__traceback__))
			return
		print(res)
		print('!!!!!!!!!!!!!!!!!!!!!!!')
		res, err = await b.encrypt_data(b'HEELLO WORLD!')
		if err is not None:
			print(err)
			print(traceback.format_tb(err.__traceback__))
			return
		print(res)

		res, err = await b.decrypt_data(res)
		if err is not None:
			print(err)
			print(traceback.format_tb(err.__traceback__))
			return
		print(res)

if __name__ == '__main__':
	url = 'smb2+ntlm-password://TEST\\victim:Passw0rd!1@10.10.10.2'
	asyncio.run(amain(url))
	


# SECUREAUTH LABS. Copyright 2018 SecureAuth Corporation. All rights reserved.
#
# This software is provided under under a slightly modified version
# of the Apache Software License. See the accompanying LICENSE file
# for more information.
#
# Author: Alberto Solino (@agsolino)
#
# Description:
#   [MS-RRP] Interface implementation
#
#   Best way to learn how to use these calls is to grab the protocol standard
#   so you understand what the call does, and then read the test case located
#   at https://github.com/SecureAuthCorp/impacket/tree/master/tests/SMB_RPC
#
#   Some calls have helper functions, which makes it even easier to use.
#   They are located at the end of this file. 
#   Helper functions start with "h"<name of the call>.
#   There are test cases for them too. 
#


from struct import unpack, pack
from aiosmb import logger as LOG

from aiosmb.dcerpc.v5.ndr import NDRCALL, NDRSTRUCT, NDRPOINTER, NDRUniConformantVaryingArray, NDRUniConformantArray
from aiosmb.dcerpc.v5.dtypes import DWORD, UUID, ULONG, LPULONG, BOOLEAN, SECURITY_INFORMATION, PFILETIME, \
	RPC_UNICODE_STRING, FILETIME, NULL, MAXIMUM_ALLOWED, OWNER_SECURITY_INFORMATION, PWCHAR, PRPC_UNICODE_STRING
from aiosmb.dcerpc.v5.rpcrt import DCERPCException
from aiosmb.dcerpc.v5.uuid import uuidtup_to_bin
from aiosmb.dcerpc.v5 import system_errors
from aiosmb.commons.utils.decorators import red, rr

MSRPC_UUID_RRP = uuidtup_to_bin(('338CD001-2244-31F1-AAAA-900038001003', '1.0'))

class DCERPCSessionError(DCERPCException):
	def __init__(self, error_string=None, error_code=None, packet=None):
		DCERPCException.__init__(self, error_string, error_code, packet)

	def __str__( self ):
		key = self.error_code
		if key in system_errors.ERROR_MESSAGES:
			error_msg_short = system_errors.ERROR_MESSAGES[key][0]
			error_msg_verbose = system_errors.ERROR_MESSAGES[key][1]
			return 'RRP SessionError: code: 0x%x - %s - %s' % (self.error_code, error_msg_short, error_msg_verbose)
		else:
			return 'RRP SessionError: unknown error code: 0x%x' % self.error_code

################################################################################
# CONSTANTS
################################################################################
# 2.2.2 PREGISTRY_SERVER_NAME
PREGISTRY_SERVER_NAME = PWCHAR

# 2.2.3 error_status_t
error_status_t = ULONG

# 2.2.5 RRP_UNICODE_STRING
RRP_UNICODE_STRING = RPC_UNICODE_STRING
PRRP_UNICODE_STRING = PRPC_UNICODE_STRING

# 2.2.4 REGSAM
REGSAM = ULONG

KEY_QUERY_VALUE        = 0x00000001
KEY_SET_VALUE          = 0x00000002
KEY_CREATE_SUB_KEY     = 0x00000004
KEY_ENUMERATE_SUB_KEYS = 0x00000008
KEY_CREATE_LINK        = 0x00000020
KEY_WOW64_64KEY        = 0x00000100
KEY_WOW64_32KEY        = 0x00000200

REG_BINARY              = 3
REG_DWORD               = 4
REG_DWORD_LITTLE_ENDIAN = 4
REG_DWORD_BIG_ENDIAN    = 5
REG_EXPAND_SZ           = 2
REG_LINK                = 6
REG_MULTI_SZ            = 7
REG_NONE                = 0
REG_QWORD               = 11
REG_QWORD_LITTLE_ENDIAN = 11
REG_SZ                  = 1 

# 3.1.5.7 BaseRegCreateKey (Opnum 6)
REG_CREATED_NEW_KEY     = 0x00000001
REG_OPENED_EXISTING_KEY = 0x00000002

# 3.1.5.19 BaseRegRestoreKey (Opnum 19)
# Flags
REG_WHOLE_HIVE_VOLATILE = 0x00000001
REG_REFRESH_HIVE        = 0x00000002
REG_NO_LAZY_FLUSH       = 0x00000004
REG_FORCE_RESTORE       = 0x00000008

################################################################################
# STRUCTURES
################################################################################
# 2.2.1 RPC_HKEY
class RPC_HKEY(NDRSTRUCT):
	structure =  (
		('context_handle_attributes',ULONG),
		('context_handle_uuid',UUID),
	)
	def __init__(self, data = None,isNDR64 = False):
		NDRSTRUCT.__init__(self, data, isNDR64)
		self['context_handle_uuid'] = '\x00'*20

# 2.2.6 RVALENT
class RVALENT(NDRSTRUCT):
	structure =  (
		('ve_valuename',PRRP_UNICODE_STRING),
		('ve_valuelen',DWORD),
		('ve_valueptr',DWORD),
		('ve_type',DWORD),
	)

class RVALENT_ARRAY(NDRUniConformantVaryingArray):
	item = RVALENT

# 2.2.9 RPC_SECURITY_DESCRIPTOR
class BYTE_ARRAY(NDRUniConformantVaryingArray):
	pass

class PBYTE_ARRAY(NDRPOINTER):
	referent = (
		('Data', BYTE_ARRAY),
	)

class RPC_SECURITY_DESCRIPTOR(NDRSTRUCT):
	structure =  (
		('lpSecurityDescriptor',PBYTE_ARRAY),
		('cbInSecurityDescriptor',DWORD),
		('cbOutSecurityDescriptor',DWORD),
	)

# 2.2.8 RPC_SECURITY_ATTRIBUTES
class RPC_SECURITY_ATTRIBUTES(NDRSTRUCT):
	structure =  (
		('nLength',DWORD),
		('RpcSecurityDescriptor',RPC_SECURITY_DESCRIPTOR),
		('bInheritHandle',BOOLEAN),
	)

class PRPC_SECURITY_ATTRIBUTES(NDRPOINTER):
	referent = (
		('Data', RPC_SECURITY_ATTRIBUTES),
	)

################################################################################
# RPC CALLS
################################################################################
# 3.1.5.1 OpenClassesRoot (Opnum 0)
class OpenClassesRoot(NDRCALL):
	opnum = 0
	structure = (
	   ('ServerName', PREGISTRY_SERVER_NAME),
	   ('samDesired', REGSAM),
	)

class OpenClassesRootResponse(NDRCALL):
	structure = (
	   ('phKey', RPC_HKEY),
	   ('ErrorCode', error_status_t),
	)

# 3.1.5.2 OpenCurrentUser (Opnum 1)
class OpenCurrentUser(NDRCALL):
	opnum = 1
	structure = (
	   ('ServerName', PREGISTRY_SERVER_NAME),
	   ('samDesired', REGSAM),
	)

class OpenCurrentUserResponse(NDRCALL):
	structure = (
	   ('phKey', RPC_HKEY),
	   ('ErrorCode', error_status_t),
	)

# 3.1.5.3 OpenLocalMachine (Opnum 2)
class OpenLocalMachine(NDRCALL):
	opnum = 2
	structure = (
	   ('ServerName', PREGISTRY_SERVER_NAME),
	   ('samDesired', REGSAM),
	)

class OpenLocalMachineResponse(NDRCALL):
	structure = (
	   ('phKey', RPC_HKEY),
	   ('ErrorCode', error_status_t),
	)

# 3.1.5.4 OpenPerformanceData (Opnum 3)
class OpenPerformanceData(NDRCALL):
	opnum = 3
	structure = (
	   ('ServerName', PREGISTRY_SERVER_NAME),
	   ('samDesired', REGSAM),
	)

class OpenPerformanceDataResponse(NDRCALL):
	structure = (
	   ('phKey', RPC_HKEY),
	   ('ErrorCode', error_status_t),
	)

# 3.1.5.5 OpenUsers (Opnum 4)
class OpenUsers(NDRCALL):
	opnum = 4
	structure = (
	   ('ServerName', PREGISTRY_SERVER_NAME),
	   ('samDesired', REGSAM),
	)

class OpenUsersResponse(NDRCALL):
	structure = (
	   ('phKey', RPC_HKEY),
	   ('ErrorCode', error_status_t),
	)

# 3.1.5.6 BaseRegCloseKey (Opnum 5)
class BaseRegCloseKey(NDRCALL):
	opnum = 5
	structure = (
	   ('hKey', RPC_HKEY),
	)

class BaseRegCloseKeyResponse(NDRCALL):
	structure = (
	   ('hKey', RPC_HKEY),
	   ('ErrorCode', error_status_t),
	)

# 3.1.5.7 BaseRegCreateKey (Opnum 6)
class BaseRegCreateKey(NDRCALL):
	opnum = 6
	structure = (
	   ('hKey', RPC_HKEY),
	   ('lpSubKey', RRP_UNICODE_STRING),
	   ('lpClass', RRP_UNICODE_STRING),
	   ('dwOptions', DWORD),
	   ('samDesired', REGSAM),
	   ('lpSecurityAttributes', PRPC_SECURITY_ATTRIBUTES),
	   ('lpdwDisposition', LPULONG),
	)

class BaseRegCreateKeyResponse(NDRCALL):
	structure = (
	   ('phkResult', RPC_HKEY),
	   ('lpdwDisposition', LPULONG),
	   ('ErrorCode', error_status_t),
	)

# 3.1.5.8 BaseRegDeleteKey (Opnum 7)
class BaseRegDeleteKey(NDRCALL):
	opnum = 7
	structure = (
	   ('hKey', RPC_HKEY),
	   ('lpSubKey', RRP_UNICODE_STRING),
	)

class BaseRegDeleteKeyResponse(NDRCALL):
	structure = (
	   ('ErrorCode', error_status_t),
	)

# 3.1.5.9 BaseRegDeleteValue (Opnum 8)
class BaseRegDeleteValue(NDRCALL):
	opnum = 8
	structure = (
	   ('hKey', RPC_HKEY),
	   ('lpValueName', RRP_UNICODE_STRING),
	)

class BaseRegDeleteValueResponse(NDRCALL):
	structure = (
	   ('ErrorCode', error_status_t),
	)

# 3.1.5.10 BaseRegEnumKey (Opnum 9)
class BaseRegEnumKey(NDRCALL):
	opnum = 9
	structure = (
	   ('hKey', RPC_HKEY),
	   ('dwIndex', DWORD),
	   ('lpNameIn', RRP_UNICODE_STRING),
	   ('lpClassIn', PRRP_UNICODE_STRING),
	   ('lpftLastWriteTime', PFILETIME),
	)

class BaseRegEnumKeyResponse(NDRCALL):
	structure = (
	   ('lpNameOut', RRP_UNICODE_STRING),
	   ('lplpClassOut', PRRP_UNICODE_STRING),
	   ('lpftLastWriteTime', PFILETIME),
	   ('ErrorCode', error_status_t),
	)

# 3.1.5.11 BaseRegEnumValue (Opnum 10)
class BaseRegEnumValue(NDRCALL):
	opnum = 10
	structure = (
	   ('hKey', RPC_HKEY),
	   ('dwIndex', DWORD),
	   ('lpValueNameIn', RRP_UNICODE_STRING),
	   ('lpType', LPULONG),
	   ('lpData', PBYTE_ARRAY),
	   ('lpcbData', LPULONG),
	   ('lpcbLen', LPULONG),
	)

class BaseRegEnumValueResponse(NDRCALL):
	structure = (
	   ('lpValueNameOut', RRP_UNICODE_STRING),
	   ('lpType', LPULONG),
	   ('lpData', PBYTE_ARRAY),
	   ('lpcbData', LPULONG),
	   ('lpcbLen', LPULONG),
	   ('ErrorCode', error_status_t),
	)

# 3.1.5.12 BaseRegFlushKey (Opnum 11)
class BaseRegFlushKey(NDRCALL):
	opnum = 11
	structure = (
	   ('hKey', RPC_HKEY),
	)

class BaseRegFlushKeyResponse(NDRCALL):
	structure = (
	   ('ErrorCode', error_status_t),
	)

# 3.1.5.13 BaseRegGetKeySecurity (Opnum 12)
class BaseRegGetKeySecurity(NDRCALL):
	opnum = 12
	structure = (
	   ('hKey', RPC_HKEY),
	   ('SecurityInformation', SECURITY_INFORMATION),
	   ('pRpcSecurityDescriptorIn', RPC_SECURITY_DESCRIPTOR),
	)

class BaseRegGetKeySecurityResponse(NDRCALL):
	structure = (
	   ('pRpcSecurityDescriptorOut', RPC_SECURITY_DESCRIPTOR),
	   ('ErrorCode', error_status_t),
	)

# 3.1.5.14 BaseRegLoadKey (Opnum 13)
class BaseRegLoadKey(NDRCALL):
	opnum = 13
	structure = (
	   ('hKey', RPC_HKEY),
	   ('lpSubKey', RRP_UNICODE_STRING),
	   ('lpFile', RRP_UNICODE_STRING),
	)

class BaseRegLoadKeyResponse(NDRCALL):
	structure = (
	   ('ErrorCode', error_status_t),
	)

# 3.1.5.15 BaseRegOpenKey (Opnum 15)
class BaseRegOpenKey(NDRCALL):
	opnum = 15
	structure = (
	   ('hKey', RPC_HKEY),
	   ('lpSubKey', RRP_UNICODE_STRING),
	   ('dwOptions', DWORD),
	   ('samDesired', REGSAM),
	)

class BaseRegOpenKeyResponse(NDRCALL):
	structure = (
	   ('phkResult', RPC_HKEY),
	   ('ErrorCode', error_status_t),
	)

# 3.1.5.16 BaseRegQueryInfoKey (Opnum 16)
class BaseRegQueryInfoKey(NDRCALL):
	opnum = 16
	structure = (
	   ('hKey', RPC_HKEY),
	   ('lpClassIn', RRP_UNICODE_STRING),
	)

class BaseRegQueryInfoKeyResponse(NDRCALL):
	structure = (
	   ('lpClassOut', RPC_UNICODE_STRING),
	   ('lpcSubKeys', DWORD),
	   ('lpcbMaxSubKeyLen', DWORD),
	   ('lpcbMaxClassLen', DWORD),
	   ('lpcValues', DWORD),
	   ('lpcbMaxValueNameLen', DWORD),
	   ('lpcbMaxValueLen', DWORD),
	   ('lpcbSecurityDescriptor', DWORD),
	   ('lpftLastWriteTime', FILETIME),
	   ('ErrorCode', error_status_t),
	)

# 3.1.5.17 BaseRegQueryValue (Opnum 17)
class BaseRegQueryValue(NDRCALL):
	opnum = 17
	structure = (
	   ('hKey', RPC_HKEY),
	   ('lpValueName', RRP_UNICODE_STRING),
	   ('lpType', LPULONG),
	   ('lpData', PBYTE_ARRAY),
	   ('lpcbData', LPULONG),
	   ('lpcbLen', LPULONG),
	)

class BaseRegQueryValueResponse(NDRCALL):
	structure = (
	   ('lpType', LPULONG),
	   ('lpData', PBYTE_ARRAY),
	   ('lpcbData', LPULONG),
	   ('lpcbLen', LPULONG),
	   ('ErrorCode', error_status_t),
	)

# 3.1.5.18 BaseRegReplaceKey (Opnum 18)
class BaseRegReplaceKey(NDRCALL):
	opnum = 18
	structure = (
	   ('hKey', RPC_HKEY),
	   ('lpSubKey', RRP_UNICODE_STRING),
	   ('lpNewFile', RRP_UNICODE_STRING),
	   ('lpOldFile', RRP_UNICODE_STRING),
	)

class BaseRegReplaceKeyResponse(NDRCALL):
	structure = (
	   ('ErrorCode', error_status_t),
	)

# 3.1.5.19 BaseRegRestoreKey (Opnum 19)
class BaseRegRestoreKey(NDRCALL):
	opnum = 19
	structure = (
	   ('hKey', RPC_HKEY),
	   ('lpFile', RRP_UNICODE_STRING),
	   ('Flags', DWORD),
	)

class BaseRegRestoreKeyResponse(NDRCALL):
	structure = (
	   ('ErrorCode', error_status_t),
	)

# 3.1.5.20 BaseRegSaveKey (Opnum 20)
class BaseRegSaveKey(NDRCALL):
	opnum = 20
	structure = (
	   ('hKey', RPC_HKEY),
	   ('lpFile', RRP_UNICODE_STRING),
	   ('pSecurityAttributes', PRPC_SECURITY_ATTRIBUTES),
	)

class BaseRegSaveKeyResponse(NDRCALL):
	structure = (
	   ('ErrorCode', error_status_t),
	)

# 3.1.5.21 BaseRegSetKeySecurity (Opnum 21)
class BaseRegSetKeySecurity(NDRCALL):
	opnum = 21
	structure = (
	   ('hKey', RPC_HKEY),
	   ('SecurityInformation', SECURITY_INFORMATION),
	   ('pRpcSecurityDescriptor', RPC_SECURITY_DESCRIPTOR),
	)

class BaseRegSetKeySecurityResponse(NDRCALL):
	structure = (
	   ('ErrorCode', error_status_t),
	)

# 3.1.5.22 BaseRegSetValue (Opnum 22)
class BaseRegSetValue(NDRCALL):
	opnum = 22
	structure = (
	   ('hKey', RPC_HKEY),
	   ('lpValueName', RRP_UNICODE_STRING),
	   ('dwType', DWORD),
	   ('lpData', NDRUniConformantArray),
	   ('cbData', DWORD),
	)

class BaseRegSetValueResponse(NDRCALL):
	structure = (
	   ('ErrorCode', error_status_t),
	)

# 3.1.5.23 BaseRegUnLoadKey (Opnum 23)
class BaseRegUnLoadKey(NDRCALL):
	opnum = 23
	structure = (
	   ('hKey', RPC_HKEY),
	   ('lpSubKey', RRP_UNICODE_STRING),
	)

class BaseRegUnLoadKeyResponse(NDRCALL):
	structure = (
	   ('ErrorCode', error_status_t),
	)

# 3.1.5.24 BaseRegGetVersion (Opnum 26)
class BaseRegGetVersion(NDRCALL):
	opnum = 26
	structure = (
	   ('hKey', RPC_HKEY),
	)

class BaseRegGetVersionResponse(NDRCALL):
	structure = (
	   ('lpdwVersion', DWORD),
	   ('ErrorCode', error_status_t),
	)

# 3.1.5.25 OpenCurrentConfig (Opnum 27)
class OpenCurrentConfig(NDRCALL):
	opnum = 27
	structure = (
	   ('ServerName', PREGISTRY_SERVER_NAME),
	   ('samDesired', REGSAM),
	)

class OpenCurrentConfigResponse(NDRCALL):
	structure = (
	   ('phKey', RPC_HKEY),
	   ('ErrorCode', error_status_t),
	)

# 3.1.5.26 BaseRegQueryMultipleValues (Opnum 29)
class BaseRegQueryMultipleValues(NDRCALL):
	opnum = 29
	structure = (
	   ('hKey', RPC_HKEY),
	   ('val_listIn', RVALENT_ARRAY),
	   ('num_vals', DWORD),
	   ('lpvalueBuf', PBYTE_ARRAY),
	   ('ldwTotsize', DWORD),
	)

class BaseRegQueryMultipleValuesResponse(NDRCALL):
	structure = (
	   ('val_listOut', RVALENT_ARRAY),
	   ('lpvalueBuf', PBYTE_ARRAY),
	   ('ldwTotsize', DWORD),
	   ('ErrorCode', error_status_t),
	)

# 3.1.5.27 BaseRegSaveKeyEx (Opnum 31)
class BaseRegSaveKeyEx(NDRCALL):
	opnum = 31
	structure = (
	   ('hKey', RPC_HKEY),
	   ('lpFile', RRP_UNICODE_STRING),
	   ('pSecurityAttributes', PRPC_SECURITY_ATTRIBUTES),
	   ('Flags', DWORD),
	)

class BaseRegSaveKeyExResponse(NDRCALL):
	structure = (
	   ('ErrorCode', error_status_t),
	)

# 3.1.5.28 OpenPerformanceText (Opnum 32)
class OpenPerformanceText(NDRCALL):
	opnum = 32
	structure = (
	   ('ServerName', PREGISTRY_SERVER_NAME),
	   ('samDesired', REGSAM),
	)

class OpenPerformanceTextResponse(NDRCALL):
	structure = (
	   ('phKey', RPC_HKEY),
	   ('ErrorCode', error_status_t),
	)

# 3.1.5.29 OpenPerformanceNlsText (Opnum 33)
class OpenPerformanceNlsText(NDRCALL):
	opnum = 33
	structure = (
	   ('ServerName', PREGISTRY_SERVER_NAME),
	   ('samDesired', REGSAM),
	)

class OpenPerformanceNlsTextResponse(NDRCALL):
	structure = (
	   ('phKey', RPC_HKEY),
	   ('ErrorCode', error_status_t),
	)

# 3.1.5.30 BaseRegQueryMultipleValues2 (Opnum 34)
class BaseRegQueryMultipleValues2(NDRCALL):
	opnum = 34
	structure = (
	   ('hKey', RPC_HKEY),
	   ('val_listIn', RVALENT_ARRAY),
	   ('num_vals', DWORD),
	   ('lpvalueBuf', PBYTE_ARRAY),
	   ('ldwTotsize', DWORD),
	)

class BaseRegQueryMultipleValues2Response(NDRCALL):
	structure = (
	   ('val_listOut', RVALENT_ARRAY),
	   ('lpvalueBuf', PBYTE_ARRAY),
	   ('ldwRequiredSize', DWORD),
	   ('ErrorCode', error_status_t),
	)

# 3.1.5.31 BaseRegDeleteKeyEx (Opnum 35)
class BaseRegDeleteKeyEx(NDRCALL):
	opnum = 35
	structure = (
	   ('hKey', RPC_HKEY),
	   ('lpSubKey', RRP_UNICODE_STRING),
	   ('AccessMask', REGSAM),
	   ('Reserved', DWORD),
	)

class BaseRegDeleteKeyExResponse(NDRCALL):
	structure = (
	   ('ErrorCode', error_status_t),
	)

################################################################################
# OPNUMs and their corresponding structures
################################################################################
OPNUMS = {
 0 : (OpenClassesRoot, OpenClassesRootResponse),
 1 : (OpenCurrentUser, OpenCurrentUserResponse),
 2 : (OpenLocalMachine, OpenLocalMachineResponse),
 3 : (OpenPerformanceData, OpenPerformanceDataResponse),
 4 : (OpenUsers, OpenUsersResponse),
 5 : (BaseRegCloseKey, BaseRegCloseKeyResponse),
 6 : (BaseRegCreateKey, BaseRegCreateKeyResponse),
 7 : (BaseRegDeleteKey, BaseRegDeleteKeyResponse),
 8 : (BaseRegDeleteValue, BaseRegDeleteValueResponse),
 9 : (BaseRegEnumKey, BaseRegEnumKeyResponse),
10 : (BaseRegEnumValue, BaseRegEnumValueResponse),
11 : (BaseRegFlushKey, BaseRegFlushKeyResponse),
12 : (BaseRegGetKeySecurity, BaseRegGetKeySecurityResponse),
13 : (BaseRegLoadKey, BaseRegLoadKeyResponse),
15 : (BaseRegOpenKey, BaseRegOpenKeyResponse),
16 : (BaseRegQueryInfoKey, BaseRegQueryInfoKeyResponse),
17 : (BaseRegQueryValue, BaseRegQueryValueResponse),
18 : (BaseRegReplaceKey, BaseRegReplaceKeyResponse),
19 : (BaseRegRestoreKey, BaseRegRestoreKeyResponse),
20 : (BaseRegSaveKey, BaseRegSaveKeyResponse),
21 : (BaseRegSetKeySecurity, BaseRegSetKeySecurityResponse),
22 : (BaseRegSetValue, BaseRegSetValueResponse),
23 : (BaseRegUnLoadKey, BaseRegUnLoadKeyResponse),
26 : (BaseRegGetVersion, BaseRegGetVersionResponse),
27 : (OpenCurrentConfig, OpenCurrentConfigResponse),
29 : (BaseRegQueryMultipleValues, BaseRegQueryMultipleValuesResponse),
31 : (BaseRegSaveKeyEx, BaseRegSaveKeyExResponse),
32 : (OpenPerformanceText, OpenPerformanceTextResponse),
33 : (OpenPerformanceNlsText, OpenPerformanceNlsTextResponse),
34 : (BaseRegQueryMultipleValues2, BaseRegQueryMultipleValues2Response),
35 : (BaseRegDeleteKeyEx, BaseRegDeleteKeyExResponse),
}

################################################################################
# HELPER FUNCTIONS
################################################################################
def checkNullString(string):
	if string == NULL:
		return string

	if string[-1:] != '\x00':
		return string + '\x00'
	else:
		return string

def packValue(valueType, value):
	if valueType == REG_DWORD:
		retData = pack('<L', value)
	elif valueType == REG_DWORD_BIG_ENDIAN:
		retData = pack('>L', value)
	elif valueType == REG_EXPAND_SZ:
		try:
			retData = value.encode('utf-16le')
		except UnicodeDecodeError:
			import sys
			retData = value.decode(sys.getfilesystemencoding()).encode('utf-16le')
	elif valueType == REG_MULTI_SZ:
		try:
			retData = value.encode('utf-16le')
		except UnicodeDecodeError:
			import sys
			retData = value.decode(sys.getfilesystemencoding()).encode('utf-16le')
	elif valueType == REG_QWORD:
		retData = pack('<Q', value)
	elif valueType == REG_QWORD_LITTLE_ENDIAN:
		retData = pack('>Q', value)
	elif valueType == REG_SZ:
		try:
			retData = value.encode('utf-16le')
		except UnicodeDecodeError:
			import sys
			retData = value.decode(sys.getfilesystemencoding()).encode('utf-16le')
	else:
		retData = value

	return retData

def unpackValue(valueType, value):
	if valueType == REG_DWORD:
		retData = unpack('<L', b''.join(value))[0]
	elif valueType == REG_DWORD_BIG_ENDIAN:
		retData = unpack('>L', ''.join(value))[0]
	elif valueType == REG_EXPAND_SZ:
		retData = b''.join(value).decode('utf-16le')
	elif valueType == REG_MULTI_SZ:
		retData = b''.join(value).decode('utf-16le')
	elif valueType == REG_QWORD:
		retData = unpack('<Q', ''.join(value))[0]
	elif valueType == REG_QWORD_LITTLE_ENDIAN:
		retData = unpack('>Q', ''.join(value))[0]
	elif valueType == REG_SZ:
		retData = b''.join(value).decode('utf-16le')
	else:
		retData = b''.join(value)

	return retData

async def hOpenClassesRoot(dce, samDesired = MAXIMUM_ALLOWED):
	request = OpenClassesRoot()
	request['ServerName'] = NULL
	request['samDesired'] = samDesired
	return await dce.request(request)

async def hOpenCurrentUser(dce, samDesired = MAXIMUM_ALLOWED):
	request = OpenCurrentUser()
	request['ServerName'] = NULL
	request['samDesired'] = samDesired
	return await dce.request(request)

async def hOpenLocalMachine(dce, samDesired = MAXIMUM_ALLOWED):
	request = OpenLocalMachine()
	request['ServerName'] = NULL
	request['samDesired'] = samDesired
	return await dce.request(request)

async def hOpenPerformanceData(dce, samDesired = MAXIMUM_ALLOWED):
	request = OpenPerformanceData()
	request['ServerName'] = NULL
	request['samDesired'] = samDesired
	return await dce.request(request)

async def hOpenUsers(dce, samDesired = MAXIMUM_ALLOWED):
	request = OpenUsers()
	request['ServerName'] = NULL
	request['samDesired'] = samDesired
	return await dce.request(request)

async def hBaseRegCloseKey(dce, hKey):
	request = BaseRegCloseKey()
	request['hKey'] = hKey
	return await dce.request(request)

async def hBaseRegCreateKey(dce, hKey, lpSubKey, lpClass = NULL, dwOptions = 0x00000001, samDesired = MAXIMUM_ALLOWED, lpSecurityAttributes = NULL, lpdwDisposition = REG_CREATED_NEW_KEY):
	request = BaseRegCreateKey()
	request['hKey'] = hKey
	request['lpSubKey'] = checkNullString(lpSubKey)
	request['lpClass'] = checkNullString(lpClass)
	request['dwOptions'] = dwOptions
	request['samDesired'] = samDesired
	if lpSecurityAttributes == NULL:
		request['lpSecurityAttributes']['RpcSecurityDescriptor']['lpSecurityDescriptor'] = NULL
	else:
		request['lpSecurityAttributes'] = lpSecurityAttributes
	request['lpdwDisposition'] = lpdwDisposition

	return await dce.request(request)

async def hBaseRegDeleteKey(dce, hKey, lpSubKey):
	request = BaseRegDeleteKey()
	request['hKey'] = hKey
	request['lpSubKey'] = checkNullString(lpSubKey)
	return await dce.request(request)

async def hBaseRegEnumKey(dce, hKey, dwIndex, lpftLastWriteTime = NULL):
	request = BaseRegEnumKey()
	request['hKey'] = hKey
	request['dwIndex'] = dwIndex
	request.fields['lpNameIn'].fields['MaximumLength'] = 1024
	request.fields['lpNameIn'].fields['Data'].fields['Data'].fields['MaximumCount'] = 1024//2
	request['lpClassIn'] = ' '* 64
	request['lpftLastWriteTime'] = lpftLastWriteTime

	return await dce.request(request)

async def hBaseRegEnumValue(dce, hKey, dwIndex, dataLen=256):
	request = BaseRegEnumValue()
	request['hKey'] = hKey
	request['dwIndex'] = dwIndex
	retries = 0

	# We need to be aware the size might not be enough, so let's catch ERROR_MORE_DATA exception
	while True:
		# Only the maximum length field of the lpValueNameIn is used to determine the buffer length to be allocated
		# by the service. Specify a string with a zero length but maximum length set to the largest buffer size
		# needed to hold the value names.
		request.fields['lpValueNameIn'].fields['MaximumLength'] = dataLen*2
		request.fields['lpValueNameIn'].fields['Data'].fields['Data'].fields['MaximumCount'] = dataLen

		request['lpData'] = b' ' * dataLen
		request['lpcbData'] = dataLen
		request['lpcbLen'] = dataLen
		resp, e = await dce.request(request)
		if e is not None:
			if isinstance(e, DCERPCSessionError):
				if retries > 1:
					return None, e
				if e.get_error_code() == system_errors.ERROR_MORE_DATA:
					# We need to adjust the size
					retries +=1
					dataLen = e.get_packet()['lpcbData']
					continue
				return None, e
			return None, e
		else:
			return resp, None


async def hBaseRegFlushKey(dce, hKey):
	request = BaseRegFlushKey()
	request['hKey'] = hKey
	return await dce.request(request)

async def hBaseRegGetKeySecurity(dce, hKey, securityInformation = OWNER_SECURITY_INFORMATION , dataLen = 1024):
	request = BaseRegGetKeySecurity()
	request['hKey'] = hKey
	request['SecurityInformation'] = securityInformation
	request['pRpcSecurityDescriptorIn']['lpSecurityDescriptor'] = NULL

	retries = 0
	while True:
		request['pRpcSecurityDescriptorIn']['cbInSecurityDescriptor'] = dataLen

		resp, e = await dce.request(request)
		if e is not None:
			if isinstance(e, DCERPCSessionError):
				if retries > 1:
					return None, None, e
				if e.get_error_code() == system_errors.ERROR_INSUFFICIENT_BUFFER:
					# We need to adjust the size
					retries +=1
					dataLen = e.get_packet()['pRpcSecurityDescriptorOut']['cbInSecurityDescriptor']+1
					continue
				return None, e
			return None, e
		return b''.join(resp['pRpcSecurityDescriptorOut']['lpSecurityDescriptor']), None

async def hBaseRegLoadKey(dce, hKey, lpSubKey, lpFile):
	request = BaseRegLoadKey()
	request['hKey'] = hKey
	request['lpSubKey'] = checkNullString(lpSubKey)
	request['lpFile'] = checkNullString(lpFile)
	return await dce.request(request)

async def hBaseRegUnLoadKey(dce, hKey, lpSubKey):
	request = BaseRegUnLoadKey()
	request['hKey'] = hKey
	request['lpSubKey'] = checkNullString(lpSubKey)
	return await dce.request(request)

async def hBaseRegOpenKey(dce, hKey, lpSubKey, dwOptions=0x00000001, samDesired = MAXIMUM_ALLOWED):
	request = BaseRegOpenKey()
	request['hKey'] = hKey
	request['lpSubKey'] = checkNullString(lpSubKey)
	request['dwOptions'] = dwOptions
	request['samDesired'] = samDesired 
	return await dce.request(request)

async def hBaseRegQueryInfoKey(dce, hKey):
	request = BaseRegQueryInfoKey()
	request['hKey'] = hKey
	# Not the cleanest way, but oh well
	# Plus, Windows XP needs MaximumCount also set
	request.fields['lpClassIn'].fields['MaximumLength'] = 1024
	request.fields['lpClassIn'].fields['Data'].fields['Data'].fields['MaximumCount'] = 1024//2
	return await dce.request(request)

async def hBaseRegQueryValue(dce, hKey, lpValueName, dataLen=512, unpack_vals = True):
	request = BaseRegQueryValue()
	request['hKey'] = hKey
	request['lpValueName'] = checkNullString(lpValueName)
	retries = 1

	# We need to be aware the size might not be enough, so let's catch ERROR_MORE_DATA exception
	while True:
		request['lpData'] =b' ' * dataLen
		request['lpcbData'] = dataLen
		request['lpcbLen'] = dataLen
		resp, e = await dce.request(request)
		if e is not None:
			if isinstance(e, DCERPCSessionError):
				if retries > 1:
					return None, None, e
				if e.get_error_code() == system_errors.ERROR_MORE_DATA:
					# We need to adjust the size
					retries +=1
					dataLen = e.get_packet()['lpcbData']
					continue
				return None, None, e
			return None, None, e
		# Returns
		# ( dataType, data )
		if unpack_vals is True:
			return resp['lpType'], unpackValue(resp['lpType'], resp['lpData']), None
		return resp['lpType'], b''.join(resp['lpData']), None

async def hBaseRegReplaceKey(dce, hKey, lpSubKey, lpNewFile, lpOldFile):
	request = BaseRegReplaceKey()
	request['hKey'] = hKey
	request['lpSubKey'] = checkNullString(lpSubKey)
	request['lpNewFile'] = checkNullString(lpNewFile)
	request['lpOldFile'] = checkNullString(lpOldFile)
	return await dce.request(request)

async def hBaseRegRestoreKey(dce, hKey, lpFile, flags=REG_REFRESH_HIVE):
	request = BaseRegRestoreKey()
	request['hKey'] = hKey
	request['lpFile'] = checkNullString(lpFile)
	request['Flags'] = flags
	return await dce.request(request)

async def hBaseRegSaveKey(dce, hKey, lpFile, pSecurityAttributes = NULL):
	request = BaseRegSaveKey()
	request['hKey'] = hKey
	request['lpFile'] = checkNullString(lpFile)
	request['pSecurityAttributes'] = pSecurityAttributes
	return await dce.request(request)

async def hBaseRegSetValue(dce, hKey, lpValueName, dwType, lpData):
	request = BaseRegSetValue()
	request['hKey'] = hKey
	request['lpValueName'] = checkNullString(lpValueName)
	request['dwType'] = dwType
	request['lpData'] = packValue(dwType,lpData)
	request['cbData'] = len(request['lpData'])
	return await dce.request(request)

async def hBaseRegGetVersion(dce, hKey):
	request = BaseRegGetVersion()
	request['hKey'] = hKey
	return await dce.request(request)

async def hOpenCurrentConfig(dce, samDesired = MAXIMUM_ALLOWED):
	request = OpenCurrentConfig()
	request['ServerName'] = NULL
	request['samDesired'] = samDesired
	return await dce.request(request)

async def hBaseRegQueryMultipleValues(dce, hKey, val_listIn):
	raise Exception('MODIFICATION ASD TEST NEEDED!!!')
	# ToDo, check the result to see whether we need to 
	# have a bigger buffer for the data to receive
	request = BaseRegQueryMultipleValues()
	request['hKey'] = hKey

	for item in  val_listIn:
		itemn = RVALENT() 
		itemn['ve_valuename'] = checkNullString(item['ValueName'])
		itemn['ve_valuelen'] = len(itemn['ve_valuename'])
		itemn['ve_valueptr'] = NULL
		itemn['ve_type'] = item['ValueType']
		request['val_listIn'].append(itemn)

	request['num_vals'] = len(request['val_listIn'])
	request['lpvalueBuf'] = list(b' '*128)
	request['ldwTotsize'] = 128

	resp = await dce.request(request)
	retVal = list()
	for item in resp['val_listOut']:
		itemn = dict()
		itemn['ValueName'] = item['ve_valuename'] 
		itemn['ValueData'] = unpackValue(item['ve_type'], resp['lpvalueBuf'][item['ve_valueptr'] : item['ve_valueptr']+item['ve_valuelen']])
		retVal.append(itemn)
 
	return retVal

async def hBaseRegSaveKeyEx(dce, hKey, lpFile, pSecurityAttributes = NULL, flags=1):
	request = BaseRegSaveKeyEx()
	request['hKey'] = hKey
	request['lpFile'] = checkNullString(lpFile)
	request['pSecurityAttributes'] = pSecurityAttributes
	request['Flags'] = flags
	return await dce.request(request)

async def hOpenPerformanceText(dce, samDesired = MAXIMUM_ALLOWED):
	request = OpenPerformanceText()
	request['ServerName'] = NULL
	request['samDesired'] = samDesired
	return await dce.request(request)

async def hOpenPerformanceNlsText(dce, samDesired = MAXIMUM_ALLOWED):
	request = OpenPerformanceNlsText()
	request['ServerName'] = NULL
	request['samDesired'] = samDesired
	return await dce.request(request)

async def hBaseRegDeleteValue(dce, hKey, lpValueName):
	request = BaseRegDeleteValue()
	request['hKey'] = hKey
	request['lpValueName'] = checkNullString(lpValueName)
	return await dce.request(request)



from aiosmb.commons.connection.proxy import SMBProxyType
from aiosmb.network.tcp import TCPSocket
from aiosmb.network.socks import SocksProxyConnection
from aiosmb.network.multiplexornetwork import MultiplexorProxyConnection



class NetworkSelector:
	def __init__(self):
		pass

	@staticmethod
	async def select(target):
		if target.proxy is None:
			return TCPSocket(target = target), None
		elif target.proxy.type in [SMBProxyType.WSNET,SMBProxyType.WSNETWS, SMBProxyType.WSNETWSS, SMBProxyType.SOCKS5, SMBProxyType.SOCKS5_SSL, SMBProxyType.SOCKS4, SMBProxyType.SOCKS4_SSL]:
			return SocksProxyConnection(target = target), None

		elif target.proxy.type in [SMBProxyType.MULTIPLEXOR, SMBProxyType.MULTIPLEXOR_SSL]:
			mpc = MultiplexorProxyConnection(target)
			socks_proxy, err = await mpc.connect()
			return socks_proxy, err

		else:
			return None, Exception('Cant select correct connection type!')



import asyncio
import enum
import uuid
import logging
import json

from aiosmb import logger
from aiosmb.examples.scancommons.targetgens import *
from aiosmb.examples.scancommons.internal import *
from aiosmb.examples.scancommons.utils import *
from aiosmb.commons.utils.univeraljson import UniversalEncoder

from aiosmb.commons.connection.url import SMBConnectionURL
from aiosmb.commons.interfaces.machine import SMBMachine
from aiosmb.protocol.common import SMB_NEGOTIATE_PROTOCOL_TEST, NegotiateDialects

from tqdm import tqdm

class SMBProtocolEnumProgressResult:
	def __init__(self, total_targets, total_finished, gens_finished, target):
		self.total_targets = total_targets
		self.total_finished = total_finished
		self.gens_finished = gens_finished
		self.target = target

class SMBProtocolEnumResultInner:
	def __init__(self, target_id, target, result, error = None, status = EnumResultStatus.RESULT):
		self.target_id = target_id
		self.target = target
		self.error = error
		self.result = result
		self.status = status

class SMBProtocolEnumData:
	def __init__(self, target_id, target, protocol, sign_en, sign_req, reply):
		self.target_id = target_id
		self.target = target
		self.protocol = protocol
		self.sign_en = sign_en
		self.sign_req = sign_req
		self.reply = reply

SMBPROTOCOLENUM_TSV_HDR = ['target', 'target_id', 'protocol', 'sigen', 'sigreq' ]


class SMBProtocolEnumResult:
	def __init__(self, obj, otype):
		self.obj = obj
		self.otype = otype

	def __str__(self):
		if self.otype == 'result':
			return '[R] %s | %s | %s | %s | %s' % (self.obj.target, self.obj.target_id, self.obj.protocol, self.obj.sign_en, self.obj.sign_req)
	
		elif self.otype == 'error':
			return '[E] %s | %s | %s' % (self.obj.target, self.obj.target_id, self.obj.error)

		elif self.otype == 'progress':
			return '[P] %s/%s | %s | %s' % (self.obj.total_targets, self.obj.total_finished, str(self.obj.gens_finished), self.obj.target)

		else:
			return '[UNK]'
	
	def to_dict(self):
		if self.otype == 'result':
			t = {}
			t['target'] = self.obj.target
			t['target_id'] = self.obj.target_id
			t['protocol'] = self.obj.protocol
			t['sigen'] = self.obj.sign_en
			t['sigreq'] = self.obj.sign_req
			return t
		return {}
	
	def to_json(self):
		dd = self.to_dict()
		return json.dumps(dd, cls = UniversalEncoder)

	def to_tsv(self, hdrs = SMBPROTOCOLENUM_TSV_HDR, separator = '\t'):
		if self.otype == 'result':
			dd = self.to_dict()
			data = [ str(dd[x]) for x in hdrs ]
			return separator.join(data)

		return ''

class SMBProtocolEnum:
	def __init__(self, worker_count = 100, timeout = 5, only_signing = False, protocols = SMB_NEGOTIATE_PROTOCOL_TEST, exclude_target = [], show_pbar = False, ext_result_q = None, output_type = 'str', out_file = None):
		self.target_gens = []
		self.timeout = timeout
		self.worker_count = worker_count
		self.task_q = None
		self.res_q = None
		self.exclude_target = exclude_target
		self.workers = []
		self.result_processing_task = None
		self.show_pbar = show_pbar
		self.ext_result_q = ext_result_q
		self.output_type = output_type
		self.out_file = out_file
		self.__gens_finished = False
		self.__total_targets = 0
		self.__total_finished = 0
		self.protocols = protocols
		self.only_signing = only_signing

	async def __executor(self, tid, target):
		try:
			for protocol in self.protocols:
				smb_mgr = SMBConnectionURL('smb2+ntlm-password://%s/?timeout=%s' % (target, self.timeout))
				connection = smb_mgr.create_connection_newtarget(target)
				res, sign_en, sign_req, rply, err = await connection.protocol_test([protocol])
				if err is not None:
					raise err
				
				er = SMBProtocolEnumResultInner(tid, target, (protocol, res, sign_en, sign_req, rply, err))
				await self.res_q.put(er)
				if self.only_signing is True:
					return
		except asyncio.CancelledError:
			return
		except Exception as e:
			await self.res_q.put(SMBProtocolEnumResultInner(tid, target, None, error = e, status = EnumResultStatus.ERROR))
		finally:
			await self.res_q.put(SMBProtocolEnumResultInner(tid, target, None, status = EnumResultStatus.FINISHED))

	async def worker(self):
		try:
			while True:
				indata = await self.task_q.get()
				if indata is None:
					return
				
				tid, target = indata
				try:
					await asyncio.wait_for(self.__executor(tid, target), timeout=10)
				except asyncio.CancelledError:
					return
				except Exception as e:
					pass
		except asyncio.CancelledError:
			return
				
		except Exception as e:
			return e

	async def result_processing(self):
		try:
			pbar = None
			if self.show_pbar is True:
				pbar = {}
				pbar['targets']         = tqdm(desc='Targets                ', unit='', position=0)
				pbar['smb1']            = tqdm(desc='SMBv1                  ', unit='', position=1)
				pbar['signenabled']     = tqdm(desc='Singing enabled        ', unit='', position=2)
				pbar['signnotenforced'] = tqdm(desc='Singning NOT enforced  ', unit='', position=3)
				pbar['connerrors']      = tqdm(desc='Conn Errors            ', unit='', position=4)
			
			out_buffer = []
			final_iter = False

			while True:
				try:
					if self.__gens_finished is True and self.show_pbar is True and pbar['targets'].total is None:
						pbar['targets'].total = self.__total_targets
						for key in pbar:
							pbar[key].refresh()
					if self.ext_result_q is not None:
						out_buffer = []

					if len(out_buffer) >= 1000 or final_iter and self.ext_result_q is None:
						out_data = ''
						if self.output_type == 'str':
							out_data = '\r\n'.join([str(x) for x in out_buffer])
						elif self.output_type == 'tsv':
							for res in out_buffer:
								t = res.to_tsv()
								if t == '':
									continue
								out_data += '%s\r\n' % res.to_tsv()
						elif self.output_type == 'json':
							for res in out_buffer:
								t = res.to_json()
								if len(t) == 2:
									continue
								out_data += '%s\r\n' % t
					
						else:
							out_data = '\r\n'.join(out_buffer)

						if self.out_file is not None:
							with open(self.out_file, 'a+', newline = '') as f:
								f.write(out_data)
						
						else:
							print(out_data)
						
						if self.show_pbar is True:
							for key in pbar:
								pbar[key].refresh()
						
						out_buffer = []
						out_data = ''
					
					if final_iter:
						asyncio.create_task(self.terminate())
						return
					
					try:
						er = await asyncio.wait_for(self.res_q.get(), timeout = 5)
					except asyncio.TimeoutError:
						if self.show_pbar is True:
							for key in pbar:
								pbar[key].refresh()

						if self.__total_finished == self.__total_targets and self.__gens_finished is True:
							final_iter = True
						continue
					
					if er.status == EnumResultStatus.FINISHED:
						self.__total_finished += 1
						if self.show_pbar is True:
							pbar['targets'].update(1)
						
						obj = SMBProtocolEnumProgressResult(self.__total_targets, self.__total_finished, self.__gens_finished, er.target)
						if self.ext_result_q is not None:
							await self.ext_result_q.put(SMBProtocolEnumResult(obj, 'progress'))
						out_buffer.append(SMBProtocolEnumResult(obj, 'progress'))
						if self.__total_finished == self.__total_targets and self.__gens_finished is True:
							final_iter = True
							continue

					elif er.status == EnumResultStatus.RESULT:
						protocol, result, sign_en, sign_req, rply, err = er.result
						if protocol == NegotiateDialects.WILDCARD:
							protocol = 'SMB1' #replacing this bc of logic in connection
						else:
							protocol = protocol.name.upper()
						if result is True:
							if sign_en is True:
								sign_en = 'E'
							else:
								sign_en = 'D'
							if sign_req is True:
								sign_req = 'REQ'
							else:
								sign_req = 'NOTREQ'
						else:
							sign_en = None
							sign_req = None
						
						if self.show_pbar is True:
							if protocol == 'SMB1':
								pbar['smb1'].update(1)
							if sign_en == 'E':
								pbar['signenabled'].update(1)
							if sign_req == 'NOTREQ':
								pbar['signnotenforced'].update(1)

						data = SMBProtocolEnumData(er.target_id, er.target, protocol, sign_en, sign_req, rply)

						if self.ext_result_q is not None:
							await self.ext_result_q.put(SMBProtocolEnumResult(data, 'result'))
						out_buffer.append(SMBProtocolEnumResult(data, 'result'))
					
					elif er.status == EnumResultStatus.ERROR:
						if self.ext_result_q is not None:
							await self.ext_result_q.put(SMBProtocolEnumResult(er, 'error'))
						if self.show_pbar is True:
							pbar['connerrors'].update(1)
						out_buffer.append(SMBProtocolEnumResult(er, 'error'))					
					
				except asyncio.CancelledError:
					return
				except Exception as e:
					print(e)

		except asyncio.CancelledError:
			return
		except Exception as e:
			logger.exception('result_processing')
			asyncio.create_task(self.terminate())

	async def terminate(self):
		for worker in self.workers:
			worker.cancel()
		if self.result_processing_task is not None:
			self.result_processing_task.cancel()

	async def setup(self):
		try:
			if self.res_q is None:
				self.res_q = asyncio.Queue(self.worker_count)
				self.result_processing_task = asyncio.create_task(self.result_processing())
			if self.task_q is None:
				self.task_q = asyncio.Queue()

			for _ in range(self.worker_count):
				self.workers.append(asyncio.create_task(self.worker()))

			return True, None
		except Exception as e:
			return None, e
	
	async def __generate_targets(self):
		try:		
			for target_gen in self.target_gens:
				async for uid, target, err in target_gen.generate():
					if err is not None:
						print('Target gen error! %s' % err)
						break
					
					if target in self.exclude_target:
						continue
					
					self.__total_targets += 1
					await self.task_q.put((uid, target))
					await asyncio.sleep(0)

			self.__gens_finished = True
		except Exception as e:
			logger.exception('targetgen')
	
	async def run(self):
		try:
			_, err = await self.setup()
			if err is not None:
				raise err
			
			gen_task = asyncio.create_task(self.__generate_targets())

			await asyncio.gather(*self.workers)
			await self.result_processing_task
			return True, None

		except Exception as e:
			logger.exception('run')
			return None, e

async def amain():
	import argparse
	import sys

	parser = argparse.ArgumentParser(description='SMB Protocol enumerator. Tells which dialects suported by the remote end')
	parser.add_argument('-v', '--verbose', action='count', default=0)
	parser.add_argument('-w', '--smb-worker-count', type=int, default=100, help='Parallell count')
	parser.add_argument('-t', '--timeout', type=int, default=50, help='Timeout for each connection')
	parser.add_argument('--signing', action='store_true', help='Only check for the singing properties. (faster)')
	parser.add_argument('-s', '--stdin', action='store_true', help='Read targets from stdin')
	parser.add_argument('--json', action='store_true', help='Output in JSON format')
	parser.add_argument('--tsv', action='store_true', help='Output in TSV format. (TAB Separated Values)')
	parser.add_argument('--progress', action='store_true', help='Show progress bar')
	parser.add_argument('-o', '--out-file', help='Output file path.')
	parser.add_argument('targets', nargs='*', help = 'Hostname or IP address or file with a list of targets')
	args = parser.parse_args()
	
	if args.verbose >=1:
		logger.setLevel(logging.INFO)

	if args.verbose > 2:
		logger.setLevel(1) #enabling deep debug
		asyncio.get_event_loop().set_debug(True)
		logging.basicConfig(level=logging.DEBUG)
	
	output_type = 'str'
	if args.json is True:
		output_type = 'json'
	if args.tsv is True:
		output_type = 'tsv'

	enumerator = SMBProtocolEnum(worker_count = args.smb_worker_count, timeout = args.timeout, only_signing = args.signing, show_pbar=args.progress, out_file=args.out_file, output_type=output_type)

	notfile = []
	if len(args.targets) == 0 and args.stdin is True:
		enumerator.target_gens.append(ListTargetGen(sys.stdin))
	else:
		for target in args.targets:
			try:
				f = open(target, 'r')
				f.close()
				enumerator.target_gens.append(FileTargetGen(target))
			except:
				notfile.append(target)
	
	if len(notfile) > 0:
		enumerator.target_gens.append(ListTargetGen(notfile))

	if len(enumerator.target_gens) == 0:
		print('[-] No suitable targets were found!')
		return
		
	await enumerator.run()
	if args.progress is False:
		print('[+] Done!')

def main():
	asyncio.run(amain())

if __name__ == '__main__':
	main()



import asyncio

from aiosmb import logger
from aiosmb.commons.connection.params import SMBConnectionParams
from aiosmb.commons.connection.url import SMBConnectionURL
from aiosmb.commons.interfaces.machine import SMBMachine
from aiosmb.dcerpc.v5.common.service import SMBServiceStatus
from aiosmb.external.aiocmd.aiocmd import aiocmd
import enum

class SMBREG_COMMAND(enum.Enum):
	READ = 'R'
	ENUMVALUE = 'EV'
	ENUMKEY = 'EK'


async def amain():
	import argparse
	import sys
	import logging
	

	parser = argparse.ArgumentParser(description='Registry manipulation via SMB')
	SMBConnectionParams.extend_parser(parser)
	parser.add_argument('-v', '--verbose', action='count', default=0)
	parser.add_argument('url', help='Connection URL base, target can be set to anything. Owerrides all parameter based connection settings! Example: "smb2+ntlm-password://TEST\\victim@test"')
	parser.add_argument('commands', nargs='*', help = 'Commands in the following format: "r:HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Control\SecurityProviders\WDigest:Negotiate"')

	args = parser.parse_args()

	if args.verbose >=1:
		logger.setLevel(logging.DEBUG)

	if args.verbose > 2:
		print('setting deepdebug')
		logger.setLevel(1) #enabling deep debug
		asyncio.get_event_loop().set_debug(True)
		logging.basicConfig(level=logging.DEBUG)
	
	commands = []
	smb_url = None
	if args.url is not None:
		smb_url = args.url
	else:
		try:
			smb_url = SMBConnectionParams.parse_args(args)
		except Exception as e:
			print('Either URL or all connection parameters must be set! Error: %s' % str(e))
			sys.exit(1)
	
	#pre-parsing commands
	for cmd in args.commands:
		c, path = cmd.split(':', 1)
		c = SMBREG_COMMAND(c.upper())
		commands.append((c, path))

	connection = SMBConnectionURL(smb_url).get_connection()
	_, err = await connection.login()
	if err is not None:
		print('Login failed! Reason: %s' % str(err))
		return
	machine = SMBMachine(connection)
	#async for srv, err in machine.list_services():
	#	if err is not None:
	#		print(err)
	#		return
	#	print(srv)
	registry_srv_status, err = await machine.check_service_status("RemoteRegistry")
	if err is not None:
		print('Check service status error! %s' % err)
		return

	if registry_srv_status != SMBServiceStatus.RUNNING:
		logger.info('RemoteRegistry is not running! Starting it now..')
		res, err = await machine.enable_service("RemoteRegistry")
		if err is not None:
			print(err)
			return
		await asyncio.sleep(5) #waiting for service to start up
	
	reg_api, err = await machine.get_regapi()
	if err is not None:
		print(err)
		return
	
	## do stuff
	for cmd, target in commands:
		if cmd == SMBREG_COMMAND.READ:
			regpath, name = target.split(':',1)
			hkey, err = await reg_api.OpenRegPath(regpath)
			if err is not None:
				print(err)
				continue
			
			val_type, value, err = await reg_api.QueryValue(hkey, name)
			if err is not None:
				print(err)
				continue
			print(value)
		
		elif cmd == SMBREG_COMMAND.ENUMVALUE:
			hkey, err = await reg_api.OpenRegPath(target)
			if err is not None:
				print(err)
				continue
			
			i=0
			while True:
				value_name, value_type, value_data, err = await reg_api.EnumValue(hkey, i)
				i+=1
				if err is not None:
					print(err)
					break
				print(value_name)
				print(value_type)
				print(value_data)
		
		elif cmd == SMBREG_COMMAND.ENUMKEY:
			hkey, err = await reg_api.OpenRegPath(target)
			if err is not None:
				print(err)
				continue
			i = 0
			while True:
				res, err = await reg_api.EnumKey(hkey, i)
				i+= 1
				if err is not None:
					print(err)
					break

				print(res)


def main():
	asyncio.run(amain())

if __name__ == '__main__':
	main()


#
#
# This is just a simple interface to the minikerberos library to support SPNEGO
# 
#
# - Hardships - 
# 1. DCERPC kerberos authentication requires a complete different approach and flags,
#    also requires mutual authentication
#
# - Links - 
# 1. Most of the idea was taken from impacket
# 2. See minikerberos library

import datetime
import copy

from minikerberos.common import *

from minikerberos.protocol.asn1_structs import AP_REP, EncAPRepPart, EncryptedData, Ticket
#from minikerberos.gssapi.gssapi import get_gssapi
from aiosmb.authentication.kerberos.gssapi import get_gssapi
from aiosmb.commons.connection.proxy import  SMBProxyType
from minikerberos.protocol.structures import ChecksumFlags
from minikerberos.protocol.encryption import Enctype, Key, _enctype_table
from minikerberos.protocol.constants import MESSAGE_TYPE
from minikerberos.aioclient import AIOKerberosClient
from minikerberos.network.aioclientsockssocket import AIOKerberosClientSocksSocket
from minikerberos.common.proxy import KerberosProxy

from aiosmb import logger

class SMBKerberos:
	def __init__(self, settings):
		self.settings = settings
		self.mode = None
		self.ccred = None
		self.target = None
		self.spn = None
		self.kc = None
		
		self.session_key = None
		self.gssapi = None
		self.iterations = 0
		self.etype = None
		self.seq_number = None
		self.from_ccache = False
	
		self.setup()
		
		
	async def sign(self, data, message_no, direction = 'init'):
		return self.gssapi.GSS_GetMIC(data, message_no)	
		
	async def encrypt(self, data, message_no):
		return self.gssapi.GSS_Wrap(data, message_no)
		
	async def decrypt(self, data, message_no, direction='init', auth_data=None):
		return self.gssapi.GSS_Unwrap(data, message_no, direction=direction, auth_data=auth_data)
		
	def setup(self):
		self.mode = self.settings.mode
		self.ccred = self.settings.ccred
		self.spn = self.settings.spn
		self.target = self.settings.target
		
	
	def get_session_key(self):
		return self.session_key.contents

	async def setup_kc(self):
		try:
			if self.target.proxy is None or isinstance(self.target.proxy, KerberosProxy) is True:
				self.kc = AIOKerberosClient(self.ccred, self.target)


			elif self.target.proxy.type in [SMBProxyType.MULTIPLEXOR, SMBProxyType.MULTIPLEXOR_SSL]:
				from aiosmb.network.multiplexornetwork import MultiplexorProxyConnection
				mpc = MultiplexorProxyConnection(self.target) 
				socks_proxy, err = await mpc.connect(is_kerberos = True)
				if err is not None:
					raise err

				self.kc = AIOKerberosClient(self.ccred, socks_proxy)

			else:
				raise Exception('Unknown proxy type %s' % self.target.proxy.type)

			return None, None
		except Exception as e:
			return None, e
		
	
	async def authenticate(self, authData, flags = None, seq_number = 0, is_rpc = False):
		try:
			
			if self.kc is None:
				_, err = await self.setup_kc()
				if err is not None:
					return None, None, err


			if self.iterations == 0:
				try:
					#check TGS first, maybe ccache already has what we need
					for target in self.ccred.ccache.list_targets():
						# just printing this to debug...
						logger.debug('CCACHE target SPN record: %s' % target)
					tgs, encpart, self.session_key = await self.kc.get_TGS(self.spn)
					
					self.from_ccache = True
				except Exception as e:
					# this is normal when no credentials stored in ccache
					#tgt = await self.kc.get_TGT(override_etype=[18])
					tgt = await self.kc.get_TGT()
					tgs, encpart, self.session_key = await self.kc.get_TGS(self.spn)
				
			ap_opts = []
			if is_rpc == True:
				if self.iterations == 0:
					ap_opts.append('mutual-required')
					flags = ChecksumFlags.GSS_C_CONF_FLAG | ChecksumFlags.GSS_C_INTEG_FLAG | ChecksumFlags.GSS_C_SEQUENCE_FLAG|\
							ChecksumFlags.GSS_C_REPLAY_FLAG | ChecksumFlags.GSS_C_MUTUAL_FLAG | ChecksumFlags.GSS_C_DCE_STYLE
					
					if self.from_ccache is False:
						apreq = self.kc.construct_apreq(tgs, encpart, self.session_key, flags = flags, seq_number = seq_number, ap_opts=ap_opts)					
					else:
						apreq = self.kc.construct_apreq_from_ticket(Ticket(tgs['ticket']).dump(), self.session_key, tgs['crealm'], tgs['cname']['name-string'][0], flags = flags, seq_number = seq_number, ap_opts = ap_opts, cb_data = None)
			
					self.iterations += 1
					return apreq, False, None
					
				else:
					#mutual authentication part here
					self.seq_number = seq_number

					aprep = AP_REP.load(authData).native
					cipher = _enctype_table[int(aprep['enc-part']['etype'])]()
					cipher_text = aprep['enc-part']['cipher']
					temp = cipher.decrypt(self.session_key, 12, cipher_text)
					
					enc_part = EncAPRepPart.load(temp).native
					cipher = _enctype_table[int(enc_part['subkey']['keytype'])]()
					
					now = datetime.datetime.now(datetime.timezone.utc)
					apreppart_data = {}
					apreppart_data['cusec'] = now.microsecond
					apreppart_data['ctime'] = now.replace(microsecond=0)
					apreppart_data['seq-number'] = enc_part['seq-number']
					#print('seq %s' % enc_part['seq-number'])
					
					apreppart_data_enc = cipher.encrypt(self.session_key, 12, EncAPRepPart(apreppart_data).dump(), None)
					
					#overriding current session key
					self.session_key = Key(cipher.enctype, enc_part['subkey']['keyvalue'])
					
					ap_rep = {}
					ap_rep['pvno'] = 5 
					ap_rep['msg-type'] = MESSAGE_TYPE.KRB_AP_REP.value
					ap_rep['enc-part'] = EncryptedData({'etype': self.session_key.enctype, 'cipher': apreppart_data_enc}) 
					
					token = AP_REP(ap_rep).dump()
					self.gssapi = get_gssapi(self.session_key)
					self.iterations += 1
					
					return token, False, None
			else:
				if self.from_ccache is False:
					apreq = self.kc.construct_apreq(tgs, encpart, self.session_key, flags = flags, seq_number = seq_number, ap_opts=ap_opts)
				else:
					apreq = self.kc.construct_apreq_from_ticket(Ticket(tgs['ticket']).dump(), self.session_key, tgs['crealm'], tgs['cname']['name-string'][0], flags = flags, seq_number = seq_number, ap_opts = ap_opts, cb_data = None)
				return apreq, False, None
		
		except Exception as e:
			return None, None, e



import asyncio
import uuid

class FileTargetGen:
	def __init__(self, filename):
		self.filename = filename

	async def run(self, target_q):
		try:
			cnt = 0
			with open(self.filename, 'r') as f:
				for line in f:
					line = line.strip()
					await target_q.put((str(uuid.uuid4()), line))
					await asyncio.sleep(0)
			return cnt, None
		except Exception as e:
			return cnt, e



##this file is present to force pure ciphers to be used
## reason: pyinstaller cant do dynamic imports
## usage: overwrite the "symmetric.py" file with tihs one

from aiosmb.crypto.AES import pureAES
from aiosmb.crypto.DES import pureDES
from aiosmb.crypto.RC4 import pureRC4
from aiosmb.crypto.pure.AES.AESCCM import aesCCMEncrypt as ae
from aiosmb.crypto.pure.AES.AESCCM import aesCCMDecrypt as ad

DES = pureDES
AES = pureAES
RC4 = pureRC4

aesCCMEncrypt = ae
aesCCMDecrypt = ad




import asyncio
import enum
import uuid
import logging
import json

from aiosmb import logger
from aiosmb.examples.scancommons.targetgens import *
from aiosmb.examples.scancommons.internal import *
from aiosmb.examples.scancommons.utils import *
from aiosmb.commons.utils.univeraljson import UniversalEncoder

from aiosmb.commons.connection.url import SMBConnectionURL
from aiosmb.commons.interfaces.machine import SMBMachine
from aiosmb.commons.interfaces.share import SMBShare
from aiosmb.dcerpc.v5.interfaces.remoteregistry import RRP
from aiosmb.dcerpc.v5.interfaces.servicemanager import SMBRemoteServieManager

from tqdm import tqdm

import traceback


class SMBAdminEnumResultInner:
	def __init__(self, target_id, target, error = None, status = EnumResultStatus.RESULT):
		self.target_id = target_id
		self.target = target
		self.error = error
		self.share = None
		self.registry = None
		self.servicemgr = None
		self.status = status
	
	

class SMBAdminEnumProgressResult:
	def __init__(self, total_targets, total_finished, gens_finished, target):
		self.total_targets = total_targets
		self.total_finished = total_finished
		self.gens_finished = gens_finished
		self.target = target

SMBADMINENUM_TSV_HDR = ['target', 'target_id', 'share', 'servicemgr', 'registry' ]


class SMBAdminEnumResult:
	def __init__(self, obj, otype):
		self.obj = obj
		self.otype = otype

	def to_dict(self):
		if self.otype == 'result':
			t = {}
			t['target'] = self.obj.target
			t['target_id'] = self.obj.target_id
			t['share'] = self.obj.share
			t['servicemgr'] = self.obj.servicemgr
			t['registry'] = self.obj.registry
			return t
		return {}
	
	def to_json(self):
		dd = self.to_dict()
		return json.dumps(dd, cls = UniversalEncoder)

	def to_tsv(self, hdrs = SMBADMINENUM_TSV_HDR, separator = '\t'):
		if self.otype == 'result':
			dd = self.to_dict()
			data = [ str(dd[x]) for x in hdrs ]
			return separator.join(data)

		return ''
	
	def __str__(self):
		if self.otype == 'result':
			t = ''
			if self.obj.share is True:
				t += '[R] %s | %s | SHARE\r\n' % (self.obj.target, self.obj.target_id)
			if self.obj.registry is True:
				t += '[R] %s | %s | REG\r\n' % (self.obj.target, self.obj.target_id)
			if self.obj.servicemgr is True:
				t += '[R] %s | %s | SRV\r\n' % (self.obj.target, self.obj.target_id)
			return t
	
		elif self.otype == 'error':
			return '[E] %s | %s | %s' % (self.obj.target, self.obj.target_id, self.obj.error)

		elif self.otype == 'progress':
			return '[P] %s/%s | %s | %s' % (self.obj.total_targets, self.obj.total_finished, str(self.obj.gens_finished), self.obj.target)

		else:
			return '[UNK]'
	


class SMBAdminCheck:
	def __init__(self, smb_url, worker_count = 100, enum_url = True, exclude_target = [], show_pbar = False, ext_result_q=None, output_type = 'str', out_file = None):
		self.target_gens = []
		self.smb_mgr = SMBConnectionURL(smb_url)
		self.worker_count = worker_count
		self.task_q = None
		self.res_q = None
		self.workers = []
		self.result_processing_task = None
		self.enum_url = enum_url
		self.exclude_target = []
		self.show_pbar = show_pbar
		self.ext_result_q = ext_result_q
		self.output_type = output_type
		self.out_file = out_file
		self.__gens_finished = False
		self.__total_targets = 0
		self.__total_finished = 0

	async def __executor(self, tid, target):
		try:
			connection = self.smb_mgr.create_connection_newtarget(target)
			async with connection:
				_, err = await connection.login()
				if err is not None:
					raise err
				
				res = SMBAdminEnumResultInner(tid, target)
				share = SMBShare(
					name = 'ADMIN$',
					fullpath = '\\\\%s\\%s' % (connection.target.get_hostname_or_ip(), 'ADMIN$')
				)
				_, err = await share.connect(connection)
				res.share = True if err is None else False

				rrp = RRP(connection)
				_, err = await rrp.connect()
				res.registry = True if err is None else False


				srvmgr = SMBRemoteServieManager(connection)
				_, err = await srvmgr.connect()
				res.servicemgr = True if err is None else False

				await self.res_q.put(res)


		except asyncio.CancelledError:
			return
		except Exception as e:
			await self.res_q.put(SMBAdminEnumResultInner(tid, target, error = e, status = EnumResultStatus.ERROR))
		finally:
			await self.res_q.put(SMBAdminEnumResultInner(tid, target, status = EnumResultStatus.FINISHED))

	async def worker(self):
		try:
			while True:
				indata = await self.task_q.get()
				if indata is None:
					return
				
				tid, target = indata
				try:
					await asyncio.wait_for(self.__executor(tid, target), timeout=10)
				except asyncio.CancelledError:
					return
				except Exception as e:
					logger.exception('worker')
					continue

		except asyncio.CancelledError:
			return
				
		except Exception as e:
			return e

	async def result_processing(self):
		try:
			pbar = None
			if self.show_pbar is True:
				pbar = {}
				pbar['targets']    = tqdm(desc='Targets        ', unit='', position=0)
				pbar['share']      = tqdm(desc='C$ Access      ', unit='', position=1)
				pbar['reg']        = tqdm(desc='Registry Access', unit='', position=2)
				pbar['svc']        = tqdm(desc='Service  Access', unit='', position=3)
				pbar['connerrors'] = tqdm(desc='Conn Errors    ', unit='', position=4)

			out_buffer = []
			final_iter = False

			while True:
				try:
					if self.__gens_finished is True and self.show_pbar is True and pbar['targets'].total is None:
						pbar['targets'].total = self.__total_targets
						for key in pbar:
							pbar[key].refresh()
					if self.ext_result_q is not None:
						out_buffer = []
					
					if len(out_buffer) >= 1000 or final_iter and self.ext_result_q is None:
						out_data = ''
						if self.output_type == 'str':
							out_data = '\r\n'.join([str(x) for x in out_buffer])
						elif self.output_type == 'tsv':
							for res in out_buffer:
								out_data += '%s\r\n' % res.to_tsv()
						elif self.output_type == 'json':
							for res in out_buffer:
								out_data += '%s\r\n' % res.to_json()
					
						else:
							out_data = '\r\n'.join(out_buffer)

						if self.out_file is not None:
							with open(self.out_file, 'a+', newline = '') as f:
								f.write(out_data)
						
						else:
							print(out_data)
						
						if self.show_pbar is True:
							for key in pbar:
								pbar[key].refresh()
						
						out_buffer = []
						out_data = ''
					
					if final_iter:
						asyncio.create_task(self.terminate())
						return
					
					try:
						er = await asyncio.wait_for(self.res_q.get(), timeout = 5)
					except asyncio.TimeoutError:
						if self.show_pbar is True:
							for key in pbar:
								pbar[key].refresh()

						if self.__total_finished == self.__total_targets and self.__gens_finished is True:
							final_iter = True
						continue
					
					if er.status == EnumResultStatus.FINISHED:
						self.__total_finished += 1
						if self.show_pbar is True:
							pbar['targets'].update(1)
						
						obj = SMBAdminEnumProgressResult(self.__total_targets, self.__total_finished, self.__gens_finished, er.target)
						if self.ext_result_q is not None:
							await self.ext_result_q.put(SMBAdminEnumResult(obj, 'progress'))
						out_buffer.append(SMBAdminEnumResult(obj, 'progress'))
						if self.__total_finished == self.__total_targets and self.__gens_finished is True:
							final_iter = True
							continue
					
					elif er.status == EnumResultStatus.RESULT:
						if self.show_pbar is True:
							if er.share is True:
								pbar['share'].update(1)
							if er.registry is True:
								pbar['reg'].update(1)
							if er.servicemgr is True:
								pbar['svc'].update(1)

						if self.ext_result_q is not None:
							await self.ext_result_q.put(SMBAdminEnumResult(er, 'result'))
						out_buffer.append(SMBAdminEnumResult(er, 'result'))
					
					elif er.status == EnumResultStatus.ERROR:
						if self.ext_result_q is not None:
							await self.ext_result_q.put(SMBAdminEnumResult(er, 'error'))
						if self.show_pbar is True:
							pbar['connerrors'].update(1)
						out_buffer.append(SMBAdminEnumResult(er, 'error'))

				except asyncio.CancelledError:
					return
				except Exception as e:
					logger.exception('result_processing inner')
					continue
		except asyncio.CancelledError:
			return
		except Exception as e:
			logger.exception('result_processing main')

	async def terminate(self):
		for worker in self.workers:
			worker.cancel()
		if self.result_processing_task is not None:
			self.result_processing_task.cancel()		

	async def setup(self):
		try:
			if self.res_q is None:
				self.res_q = asyncio.Queue(self.worker_count)
				self.result_processing_task = asyncio.create_task(self.result_processing())
			if self.task_q is None:
				self.task_q = asyncio.Queue()

			for _ in range(self.worker_count):
				self.workers.append(asyncio.create_task(self.worker()))

			return True, None
		except Exception as e:
			return None, e
	
	async def __generate_targets(self):
		if self.enum_url is True:
			self.__total_targets += 1
			await self.task_q.put((str(uuid.uuid4()), self.smb_mgr.get_target().get_hostname_or_ip()))
			
		for target_gen in self.target_gens:
			async for uid, target, err in target_gen.generate():
				if err is not None:
					print('Target gen error! %s' % err)
					break
				
				if target in self.exclude_target:
					continue
				
				self.__total_targets += 1
				await self.task_q.put((uid, target))
				await asyncio.sleep(0)

		self.__gens_finished = True
	
	async def run(self):
		try:
			_, err = await self.setup()
			if err is not None:
				raise err

			gen_task = asyncio.create_task(self.__generate_targets())
			
			await asyncio.gather(*self.workers)
			await self.result_processing_task
			return True, None
		except Exception as e:
			print(e)
			return None, e

async def amain():
	import argparse
	import sys
	from aiosmb.commons.connection.params import SMBConnectionParams

	epilog = """
Output legend:
    [SHARE] C$ is accessible
    [SRV] Remote Service Manager is accessible
    [REG] Remote registry is accessible
    [E] Error
    [P] Progress (current/total)
"""

	parser = argparse.ArgumentParser(description='SMB Share enumerator', formatter_class=argparse.RawDescriptionHelpFormatter, epilog=epilog)
	SMBConnectionParams.extend_parser(parser)
	parser.add_argument('-v', '--verbose', action='count', default=0)
	parser.add_argument('-w', '--smb-worker-count', type=int, default=100, help='Parallell count')
	parser.add_argument('-o', '--out-file', help='Output file path.')
	parser.add_argument('-s', '--stdin', action='store_true', help='Read targets from stdin')
	parser.add_argument('--url', help='Connection URL base, target can be set to anything. Owerrides all parameter based connection settings! Example: "smb2+ntlm-password://TEST\\victim@test"')
	parser.add_argument('--json', action='store_true', help='Output in JSON format')
	parser.add_argument('--tsv', action='store_true', help='Output in TSV format. (TAB Separated Values)')
	parser.add_argument('--progress', action='store_true', help='Show progress bar')
	parser.add_argument('targets', nargs='*', help = 'Hostname or IP address or file with a list of targets')
	args = parser.parse_args()

	if args.verbose >=1:
		logger.setLevel(logging.DEBUG)

	if args.verbose > 2:
		print('setting deepdebug')
		logger.setLevel(1) #enabling deep debug
		asyncio.get_event_loop().set_debug(True)
		logging.basicConfig(level=logging.DEBUG)

	smb_url = None
	if args.url is not None:
		smb_url = args.url
	else:
		try:
			smb_url = SMBConnectionParams.parse_args(args)
		except Exception as e:
			print('Either URL or all connection parameters must be set! Error: %s' % str(e))
			sys.exit(1)

	output_type = 'str'
	if args.json is True:
		output_type = 'json'
	if args.tsv is True:
		output_type = 'tsv'
	
	enumerator = SMBAdminCheck(smb_url, worker_count = args.smb_worker_count, output_type=output_type, out_file=args.out_file, show_pbar = args.progress)
	
	notfile = []
	if len(args.targets) == 0 and args.stdin is True:
		enumerator.target_gens.append(ListTargetGen(sys.stdin))
	else:
		for target in args.targets:
			try:
				f = open(target, 'r')
				f.close()
				enumerator.target_gens.append(FileTargetGen(target))
			except:
				notfile.append(target)
	
	if len(notfile) > 0:
		enumerator.target_gens.append(ListTargetGen(notfile))

	if len(enumerator.target_gens) == 0:
		enumerator.enum_url = True
		
	await enumerator.run()

def main():
	asyncio.run(amain())

if __name__ == '__main__':
	main()